# JoinActors: Fair Join Pattern Matching for Actors (Artifact)

This artifact supports the theory and experiments of _Fair Join Pattern Matching
for Actors_ presented in the accompanying paper. We have implemented a Scala 3
library for actors with join patterns, including both a “brute-force” and a
tree-based stateful implementation of our deterministic fair matching semantics.
Our library can be used with the off-the-shelf Scala 3 compiler, and leverages
macros to provide an intuitive API.

## Resources Required to Run the Artifact

- OS and resource (CPU, memory, disk, GPU) used by the authors to run the
  artifact: We tested the artifact on an 11th Gen Intel Core i7-1185G7 @ 8x
  4.8GHz and 32GB of memory, running  64-bit Linux 6.8.0. The plots in the paper
  were generated on a dual Xeon E5-2687W 8-core 3.10GHz processor and 128GB of
  memory, running 64-bit Linux 5.10.27.

- Estimation of the required hardware resources for evaluation: a recent Intel
  Core 17-based computer should complete the kick-the-tires benchmarks in 15-20
  minutes. The full benchmark may require 3-4 hours on a Xeon-based system.

## Quick-start guide

To quickly evaluate the artifact, you will need to install the `just` task
runner by following the instructions at <https://github.com/casey/just>.
Besides `just`, you will need to have installed Docker (or Podman with Docker
CLI emulation).

Then, you should run the following command from the root of the artifact
directory:

```bash
just docker
```

You should see messages similar to the following:

  ```bash
    $ just docker
    docker build -t ecoop24_artifact .
    [... many lines omitted ...]
    docker run -it --rm -v $(pwd):/ecoop24_artifact -w /ecoop24_artifact ecoop24_artifact /bin/bash
    root@7a556e20f47f:/ecoop24_artifact# 
  ```

The last prompt above means that you are logged in the Docker container as root.
From there, you should execute.

```bash
just quick-run-benchmarks
```

This command should compile the code, run a reduced version of the benchmarks,
and generate the figures. It should take around 15-20 minutes on a computer with
a recent Intel Core i7 CPU. The generated figures will be saved in the
`experiment_results/plots` directory as pdf files and can be opened with a PDF
viewer.

## Overview: What does the artifact comprise?

The artifact contains source code for the join patterns library implementation (including the benchmarks) and a directory that contains the Python scripts to reproduce the figures in the paper. The source code for the join patterns library is organized as follows:

- [join-actors](join-actors): Contains the core implementation of the JoinActors library (including the benchmarks).
- [experiment_results](experiment_results): Contains the Python scripts to reproduce the figures in the paper.

- [evrete-smarthouse](evrete-smarthouse): Contains the source code for the comparison with the Evrete library.


Furthermore, the artifact contains the following Python scripts to reproduce the
figures in the paper (with the same data used in the paper, i.e. experiments run
on our machines). Note that the scripts should be run from the `experiment_results` directory:

| Figure  | Script  |
|---|---|
| Figure 1 on page 18  | `./show_figure_1.sh`  |
| Figure 2 on page 20  | `./show_figure_2.sh`  |
| Figure 3 on page 22  | `./show_figure_3.sh`  |
| Figure 4 on page 23  | `./show_figure_4.sh`  |

After running the scripts, the figures will be be saved in the `experiment_results/plots` directory as pdf
files and can be opened with a PDF viewer.

<!-- Please list for each distinct component of your artifact: -->

<!-- * type of artifact (data, code, proof, etc.)
* format (e.g., CSV, source code, binary, etc.)
* location in the container/VM -->


## Reproducing the Evaluation

To reproduce the performance evaluation described in Section 5 of the paper, please follow these steps
(after the "Kick-the-tires" instructions above).

1. **Starting the docker container**:
   From the root of the artifact directory, run the following command:

     ```bash
      just docker
     ```

    After running the command, the docker container will be started and you will be logged in as the `root` user. You should see messages similar to the following:

    ```bash
      $ just docker
      docker build -t ecoop24_artifact .
      [... many lines omitted ...]
      docker run -it --rm -v $(pwd):/ecoop24_artifact -w /ecoop24_artifact ecoop24_artifact /bin/bash
      root@7a556e20f47f:/ecoop24_artifact# 
    ```

    The container is now running and you are in the root directory of the artifact.

2. **Running the benchmarks**:
    To re-run the benchmarks with same parameters used in the paper, run the following command:

    ```bash
    just run-benchmarks
    ```

    The benchmarks will be run and the results will be saved in the
    `join-actors/benchmarks/data` directory. All benchmarks have a warmup
    phase where the JVM is allowed to warm up before the actual benchmarking
    starts. The measurements during the warmup phase are not included in the
    final results. The benchmark should take 3-4 hours on a Xeon CPU. (If this
    time is excessive, you can refer to the Kick-the-tires benchmarks above,
    which should produce results similar to those in the paper.)

3. **Reproducing the figures**: After running the `just run-benchmarks` command,
    the figures will be saved in the `experiment_results/plots` directory as pdf
    files. 

### Reproducing Individual Claims

The paper contains examples showing the use of our join patterns library, some
code snippets, and benchmarks evaluating its performance.

* Listing 1 (page 2) can be found under 
  `join-actors/core/src/main/scala/examples/FactorySimpl.scala`. The file
  is compiled correctly as part of the build.

* Listing 2 (page 16) is a fragment of the actor implementation available in
  `join-actors/core/src/main/scala/actor/Actor.scala`

<!-- **NOTE:** there is a slight difference between the actor API shown in the paper
and in the artifact: the paper uses `Continue` whereas the artifact uses
`Next()`. The final published version of the artifact will use `Continue` (as in
the paper). -->

* The plot in Figure 1 (page 19): The plots can be reproduced by running the
  following command from the `experiment_results` directory:

  ```bash
    ./show_figure_1.sh
  ```

  The plots can be found here
  `experiment_results/plots/SizeWithGuardBenchmark_Figure_2_top.pdf` and
  `experiment_results/plots/SizeWithoutGuardBenchmark_WithNoise_Figure_1_bottom.pdf`.

* The plot in Figure 2 (page 20) The plots can be reproduced by running the
  following command from the `experiment_results` directory:

  ```bash
    ./show_figure_2.sh
  ```

  The plot can be found here
  `experiment_results/plots/SizeWithGuardBenchmark_Figure_2_top.pdf`,
  `experiment_results/plots/SizeWithGuardWithNoiseBenchmark_Figure_2_middle.pdf`
  and `experiment_results/plots/SizeWithGuardWithNonMatchingPayloadBenchmark_Figure_2_bottom.pdf`.

* The code snippet on page 21 can be found in
  `join-actors/core/src/main/scala/examples/SmartHouse.scala`
  and in `join-actors/benchmarks/src/main/scala/SmartHouse.scala`

* The plot in Figure 3 (page 22): The claims with regards to the matches per
  second for the Smart House example can be reproduced by running the following
  command from the `experiment_results` directory:

  ```bash
    ./show_figure_3.sh
  ```

  The plot can be foun here
  `experiment_results/plots/SmartHouseBenchmark_Figure_3.pdf`.

* The plot in Figure 4 (page 23): The plot can be reproduced by running the
  following command from the `experiment_results` directory:

  ```bash
    ./show_figure_4.sh
  ```

  The plot can be found here
  `experiment_results/plots/BoundedBufferBenchmark_Figure_4.pdf`.

* The plots in Figure 6 (page 25) can be reproduced by running the following
  command from the `ecoop24_artifact` directory:

  ```bash
    just rete-tree-comps
  ```

  The plots can then be found in the `experiment_results/rete_tree_comps` directory.

## Details about the Artifact Reusability

Besides using the provided Docker container, our artifact can be compiled by
installing a [Java Development Kit (version 21 or later)](https://jdk.java.net/21/)
and [sbt (version 1.9 or later)](https://www.scala-sbt.org/) and running `sbt compile`. Then, `sbt` will download
the required dependencies (including the Scala 3 compiler).

The benchmarks are open source and included in the artifact source code.

The API of the library is quite simple and can be easily used in other use
cases. The general template for writing join patterns using our library is as
follows:

```scala
Actor[MessageType, ReturnType] {
  receive {
    (self : ActorRef[MessageType]) =>
      {
        case pattern1 if guard1 => action1
        case pattern2 if guard2 => action2
        ...
      }
}
```

Several examples are provided in the [examples](join-actors/core/src/main/scala/examples) directory. See [README.md](join-actors/README.md) for further details.

Our join patterns library can be extended to support additional join patterns
matching algorithms by implementing the [Matcher](join-actors/core/src/main/scala/join_patterns/Matcher.scala) trait. The same file also contains the trait
implementations `BruteForceAlgorithm` and `StatefulTreeBasedAlgorithm` discussed
in the paper.
