import argparse

import pandas as pd
import matplotlib.pyplot as plt

parser = argparse.ArgumentParser(
    description="Benchmark results plotter for RETE and Tree-based Algorithms"
)

parser.add_argument(
    "--ReteAlgoFile",
    "-r",
    metavar="rete",
    type=str,
    help="the path to data file from the RETE algorithm",
)

parser.add_argument(
    "--TreeAlgoFile",
    "-t",
    metavar="tree",
    type=str,
    help="the path to data file from the tree based algorithm",
)

parser.add_argument(
    "--WithHeavyGuards",
    "-g",
    type=bool,
    help="whether to plot the results with heavy guards",
)

args = parser.parse_args()

tree_algo_path = args.TreeAlgoFile
rete_path = args.ReteAlgoFile

x_ticks = list(range(0, 21, 4))

tree_algo_data = (
    pd.read_csv(tree_algo_path, index_col=0, header=None, delimiter=";").T / 1000
)

rete_algo_data = (
    pd.read_csv(rete_path, index_col=0, header=None, delimiter=";").T / 1000
)


N_PREFIX_MESSAGES = x_ticks[-1]
(N_ITER, N) = tree_algo_data.shape


def plot_rete_comparison(tree_data, rete_data, x_ticks, x_label, save_as, title):
    fig, ax = plt.subplots()
    plt.yscale("log")

    tree_mean = tree_data.mean()
    rete_mean = rete_data.mean()

    tree_mean.index = pd.Index(x_ticks)
    rete_mean.index = pd.Index(x_ticks)

    ax.errorbar(
        tree_mean.index,
        tree_mean,
        fmt="-o",
        label="Stateful Tree-based Algorithm",
        color="green",
    )

    ax.errorbar(
        rete_mean.index,
        rete_mean,
        fmt="-x",
        label="Evrete-based implementation",
        color="blue",
    )

    ax.boxplot(
        tree_data,
        positions=tree_mean.index,
        widths=0.8,
        patch_artist=True,
        boxprops=dict(facecolor="lightgreen"),
    )

    ax.boxplot(
        rete_data,
        positions=rete_mean.index,
        widths=0.8,
        patch_artist=True,
        boxprops=dict(facecolor="lightblue"),
    )

    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel("Time (s) - log scale")
    ax.legend()
    # ax.grid()
    plt.tight_layout()
    print(f"Saving plot as {save_as}")
    fig.savefig(save_as + ".pdf")
    plt.show()

if args.WithHeavyGuards:
    plot_rete_comparison(
        tree_algo_data,
        rete_algo_data,
        x_ticks,
        "Number of prefix messages",
        "./rete_tree_comps/Evrete_vs_Tree_With_Heavy_Guards",
        f"RETE vs. Stateful Tree-based Algorithm with {N_PREFIX_MESSAGES} Prefix Messages\nWith Heavy Guards",
    )
else:
    plot_rete_comparison(
        tree_algo_data,
        rete_algo_data,
        x_ticks,
        "Number of prefix messages",
        "./rete_tree_comps/Evrete_vs_Tree_Without_Heavy_Guards",
        f"RETE vs. Stateful Tree-based Algorithm with {N_PREFIX_MESSAGES} Prefix Messages",
    )
