import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def set_size(width, fraction=1):
    """Set figure dimensions to avoid scaling in LaTeX.

    Parameters
    ----------
    width: float
            Document textwidth or columnwidth in pts
    fraction: float, optional
            Fraction of the width which you wish the figure to occupy

    Returns
    -------
    fig_dim: tuple
            Dimensions of figure in inches

    References
    ----------
    [1] https://jwalton.info/Embed-Publication-Matplotlib-Latex/
    """
    # Width of figure (in pts)
    fig_width_pt = width * fraction

    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Golden ratio to set aesthetic figure height
    # https://disq.us/p/2940ij3
    golden_ratio = (5**.5 - 1) / 2

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt
    # Figure height in inches
    fig_height_in = fig_width_in * golden_ratio

    fig_dim = (fig_width_in, fig_height_in)

    return fig_dim


plt.style.available
# plt.style.use('seaborn')

# Set the style of the plots to the seaborn style (pretty)
sns.set_theme(style="whitegrid")
plt.rcParams["figure.figsize"] = set_size(width=365)
plt.rcParams["figure.autolayout"] = True
# set dpi
plt.rcParams["figure.dpi"] = 150

tex_fonts = {
    # Use LaTeX to write all text
    "text.usetex": True,
    'font.family' : 'lmodern',
    # Use 10pt font in plots, to match 10pt font in document
    "axes.labelsize": 10,
    "font.size": 10,
    # Make the legend/label fonts a little smaller
    "legend.fontsize": 8,
    "xtick.labelsize": 8,
    "ytick.labelsize": 8,
    "axes.titlesize": 10,
}

plt.rcParams.update(tex_fonts)

brute_data1 = pd.read_csv("./brute_force_matcher_size_0_random_msgs_3_valid_msgs.csv", index_col=0)
brute_data2 = pd.read_csv("./brute_force_matcher_size_3_random_msgs_3_valid_msgs.csv", index_col=0)
brute_data3 = pd.read_csv("./brute_force_matcher_size_6_random_msgs_3_valid_msgs.csv", index_col=0)

tree_data1 = pd.read_csv("./stateful_tree_matcher_mtree_size_0_random_msgs_3_valid_msgs_V4.csv", index_col=0)
tree_data2 = pd.read_csv("./stateful_tree_matcher_mtree_size_3_random_msgs_3_valid_msgs_V4.csv", index_col=0)
tree_data3 = pd.read_csv("./stateful_tree_matcher_mtree_size_6_random_msgs_3_valid_msgs_V4.csv", index_col=0)


def plot_data(brute_data, tree_data, title, save_as, legend_loc=['upper left', 'lower right']):
    fig, ax = plt.subplots()
    ax.plot(brute_data.index, brute_data.loc[:, 'Mailbox Size'], label='Mailbox Size', color='red')
    ax.set_ylabel('Mailbox Size')
    ax.legend(loc=legend_loc[0])
    ax2 = ax.twinx()
    ax2.plot(brute_data.index, tree_data.loc[:, 'Matching Trees Size'], label='Match Tree Size', color='green', linestyle='dashed')
    ax2.set_ylabel('Number of Nodes in M-Tree')
    fig.tight_layout()
    ax2.legend(loc=legend_loc[1])
    ax.set_xlabel('Number of Messages Processed')
    plt.title(title)
    print(f"Saved as {save_as}.pdf")
    fig.savefig(save_as + '.pdf')
    # plt.show()

def plot_correlation(brute_data, tree_data, title, save_as, x_ticks=None):
    fig, ax = plt.subplots()
    ax.scatter(brute_data.loc[:, 'Mailbox Size'], tree_data.loc[:, 'Matching Trees Size'], c='red', s=8)
    # ax.set_xticks(x_ticks)
    ax.set_xlabel('Mailbox Size')
    ax.set_ylabel('Number of Nodes in M-Tree')
    plt.title(title)
    fig.tight_layout()
    print(f"Saved as {save_as}.pdf")
    fig.savefig(save_as + '.pdf')
    # plt.show()


plot_correlation(brute_data1,
                 tree_data1,
                 'Correlation Between Mailbox Size and Matching Trees Size\nWithout Random Messages Per Match',
                 'correlation_mailbox_size_vs_matching_tree_size_with_0_random_msgs_per_match',
                 x_ticks=[0, 1, 2, 3]
                 )

plot_correlation(brute_data2, 
                 tree_data2, 
                 'Correlation Between Mailbox Size and Matching Trees Size\nWith 3 Random Messages Per Match', 
                 'correlation_mailbox_size_vs_matching_tree_size_with_3_random_msgs_per_match')

plot_correlation(brute_data3,
                 tree_data3,
                 'Correlation Between Mailbox Size and Matching Trees Size\nWith 6 Random Messages Per Match',
                 'correlation_mailbox_size_vs_matching_tree_size_with_6_random_msgs_per_match')

plot_data(brute_data1, 
          tree_data1, 
          'Mailbox Size vs Matching Trees Size Without Random Messages Per Match', 
          'mailbox_size_vs_matching_tree_size_with_0_random_msgs_per_match',
          legend_loc=['lower center', 'lower right'])

plot_data(brute_data2, 
          tree_data2, 
          'Mailbox Size vs Matching Trees Size With 3 Random Messages Per Match',
          'mailbox_size_vs_matching_tree_size_with_3_random_msgs_per_match',
          legend_loc=['upper left', 'lower right'])

plot_data(brute_data3, 
          tree_data3, 
          'Mailbox Size vs Matching Trees Size With 6 Random Messages Per Match', 
          'mailbox_size_vs_matching_tree_size_with_6_random_msgs_per_match',
          legend_loc=['upper left', 'lower right'])

